# Copied from dedicated project to avoid additional dependency:
#   https://github.com/trim21/wcwidth-stubs/blob/master/src/wcwidth-stubs/__init__.pyi
def list_versions() -> tuple[str, ...]: ...
def wcwidth(wc: str, unicode_version: str = ...) -> int: ...
def wcswidth(
    pwcs: str, n: int | None = None, unicode_version: str = ...
) -> int: ...

__all__ = ("list_versions", "wcswidth", "wcwidth")
